/* treetopgm - make a listing from `find' into a graphical tree.
 * Copyright (C) 1996,8 Russell Marks.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 * treetopgm.c - main program. In fact, just about everything. :-)
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "font.h"
#include "writemrf.h"


/* types for tree structure created by parsetree.
 *
 * this is a bit awkward, as we have to allow any number of branches
 * at each level.
 */

/* both 't' and 'kids' are malloc()ed, as are the child nodes
 * themselves, of course.
 */
struct node_tag
  {
  char *t;	/* text of name */
  struct node_tag **kids;	/* array of ptrs to children */
  int knum;			/* number of children */
  int ksiz;			/* size of kids array, in bytes */
  
  int w,h;		/* width/height of text box for this node */
  int cw,ch;		/* width/height of combined child subtrees,
  			   i.e. entire subtree apart from this node */
  int sw,sh;		/* width/height of this entire subtree */
  };


struct node_tag root;	/* of tree */


/* now, stuff for the pgm itself.
 *
 * we use a font defined in font.c which has a character cell size of
 * FONT_XDIM by FONT_YDIM. The names are drawn with this. Each name
 * has a box around it, leaving a margin of BOX_MARGIN empty pixels
 * all around the name. This makes the box 2+BOX_MARGIN*2+FONT_YDIM high,
 * and 2+BOX_MARGIN*2+FONT_XDIM*(num_chars) wide.
 *
 * Each box is connected by a line from the middle of the top side to
 * the middle of the bottom side of its parent.
 *
 * The horizontal space between children, groups of children, etc. is
 * CHILD_HORIZ_GAP. The vertical space between parent/child is similarly
 * CHILD_VERT_GAP.
 *
 * Finally, there is a blank margin of BORDER_SIZE around the whole tree.
 * Due to possible rounding error when positioning, BORDER_SIZE must be >=1.
 *
 * Eventually most of these should be cmdline options, I s'pose. For now
 * they're just #defined. (XXX)
 */

#define BOX_MARGIN		(FONT_YDIM/2)
/* generally the vertical gap should be larger than the horiz. gap */
#define CHILD_HORIZ_GAP		10
#define CHILD_VERT_GAP		40
#define BORDER_SIZE		15

/* a few more based on the above. */
#define BOX_WIDTH(chars)	(2+2*BOX_MARGIN+FONT_XDIM*(chars))
#define BOX_HEIGHT		(2+2*BOX_MARGIN+FONT_YDIM)

/* after all the sizes are calculated by treegrok, it sets these: */
int pic_width,pic_height;

/* and writepgm uses this to construct the pgm before writing: */
unsigned char *pic;

/* these set whether or not to actually output tga/mrf rather than pgm: */
int output_tga=0;
int output_mrf=0;

/* these hold the grey levels to use for each part of the diagram.
 * they're in the range 0..255.
 * Note that the mrf output routine converts <128 to black and >=128
 * to white; bear this in mind if you change the shades.
 */

int grey_bg=255;		/* background */
int grey_textfg=0;		/* text */
int grey_textbg=224;		/* inside of box - background to text */
int grey_boxline=0;		/* the box outline */
int grey_line=0;		/* the line connecting parent/child */

/* XXX perhaps add optional shadow? */



void *xmalloc(int size)
{
void *ptr=malloc(size);

if(ptr==NULL) fprintf(stderr,"treetopgm: out of memory!\n"),exit(1);
return(ptr);
}


void *xrealloc(void *ptr,int size)
{
ptr=realloc(ptr,size);

if(ptr==NULL) fprintf(stderr,"treetopgm: out of memory!\n"),exit(1);
return(ptr);
}


/* create new child node and return pointer to it. Note that
 * new node is NOT initialised! You must call initnode afterwards to do
 * that.
 */
struct node_tag *newnode(struct node_tag *parent)
{
struct node_tag **arrptr,*newptr;
int reqsiz;	/* min. array size needed (in bytes) */
int incrnum=4;	/* add space for this many new entries if array too small */

/* make parent's kids array bigger if needed */
parent->knum++;
reqsiz=sizeof(struct node_tag)*parent->knum;
if(reqsiz > parent->ksiz)
  {
  /* this is overly defensive, really... */
  while(parent->ksiz < reqsiz)
    parent->ksiz+=incrnum*sizeof(struct node_tag);
  
  parent->kids=xrealloc(parent->kids,parent->ksiz);
  }

arrptr=parent->kids+parent->knum-1;

/* alloc new node */
newptr=xmalloc(sizeof(struct node_tag));
*arrptr=newptr;

return(newptr);
}


void initnode(struct node_tag *nptr,char *text)
{
nptr->t=xmalloc(strlen(text)+1);
strcpy(nptr->t,text);

/* also malloc a small initial kids array (in case we're having to use
 * a buggy realloc which can't handle a NULL arg, etc.) Don't change this
 * to init as NULL though; some other code depends on having *something*
 * malloc()ed to start with.
 */
nptr->ksiz=sizeof(struct node_tag);
nptr->kids=xmalloc(nptr->ksiz);
nptr->knum=0;			/* 0 children, though! */
}


/* return ptr to child node with text given, or NULL if none match. */
struct node_tag *childptr(struct node_tag *nptr,char *text)
{
struct node_tag **cptr=nptr->kids;
int f;

for(f=0;f<nptr->knum;f++,cptr++)
  if(strcmp((*cptr)->t,text)==0)
    return(*cptr);

return(NULL);
}


/* chop any trailing LF */
void chop(char *str)
{
if(*str && str[strlen(str)-1]=='\n')
  str[strlen(str)-1]=0;
}


/* parse 'find . -type d' tree on stdin; should also work for a plain
 * 'find', but files aren't treated specially, so that could get a bit
 * confusing.
 */
void parsetree()
{
char skip[1024],buf[1024],*ptr;
struct node_tag *nptr,*nptrnew;
int f,len;

/* treat first line as top node, and as text to be stripped from remaining
 * lines before further parsing.
 */
if(fgets(skip,sizeof(skip),stdin)==NULL)
  fprintf(stderr,"treetopgm: no input to parse!\n"),exit(1);

chop(skip);	/* chop LF */

/* also remove a trailing '/'s if that's not all there is. */
if(strlen(skip)>1 && skip[strlen(skip)-1]=='/') skip[strlen(skip)-1]=0;

/* setup top node */
initnode(&root,skip);

/* the '-1' is here for a reason - see below */
while(fgets(buf,sizeof(buf)-1,stdin)!=NULL)
  {
  chop(buf);
  ptr=buf;
  /* skip path to top node */
  if(strncmp(buf,skip,strlen(skip))==0) ptr+=strlen(skip);
  
  /* make it N asciiz strings, terminated by a zero-length one;
   * this is why we read sizeof(buf)-1 in the fgets, as we need
   * to reserve an extra char for this terminating 0-len string.
   * so, turn '/'s to NULs and add a NUL.
   */
  len=strlen(ptr);
  /* add the NUL first... */
  ptr[strlen(ptr)+1]=0;
  /* now convert the '/'s. */
  for(f=0;f<len;f++)
    if(ptr[f]=='/') ptr[f]=0;
  
  /* we should be pointing to the first '/', now a NUL - skip it.
   * (not the case if skip=="/", so test it first...)
   */
  if(*ptr==0) ptr++;
  
  /* now we go through each string in turn, creating a new node if
   * needed.
   */
  
  nptr=&root;
  while(*ptr)
    {
    if((nptrnew=childptr(nptr,ptr))!=NULL)
      nptr=nptrnew;
    else
      {
      /* must create the node first. */
      nptr=newnode(nptr);
      initnode(nptr,ptr);
      }
    
    ptr+=strlen(ptr)+1;	/* point at next string */
    }
  }
}


void treegrok_main(struct node_tag *nptr,int *wp,int *hp)
{
int f,w,h,wtmp,htmp;

/* set our own width/height */
nptr->w=BOX_WIDTH(strlen(nptr->t));
nptr->h=BOX_HEIGHT;

/* work out this entire subtree's width/height */
w=h=0;
for(f=0;f<nptr->knum;f++)
  {
  treegrok_main(*(nptr->kids+f),&wtmp,&htmp);
  /* get the total width, but only the largest height */
  w+=wtmp;
  if(htmp>h) h=htmp;
  }

/* add in the horizontal gaps between children */
if(nptr->knum>1) w+=(nptr->knum-1)*CHILD_HORIZ_GAP;

/* now w/h are the total w/h of the subtree apart from us, the root node.
 * put this in nptr->cw/nptr->ch.
 */
nptr->cw=w; nptr->ch=h;

/* add us into the tree size to calc entire subtree w/h.
 * if there are zero subnodes, use our box alone for w/h,
 * else, h+=nptr->h+CHILD_VERT_GAP, and if nptr->w > w, w=nptr->w.
 *
 * after all this, put w/h in nptr->sw/nptr->sh.
 */

if(nptr->knum==0)
  w=nptr->w,h=nptr->h;
else
  {
  h+=nptr->h+CHILD_VERT_GAP;
  if(nptr->w > w) w=nptr->w;
  }

nptr->sw=w; nptr->sh=h;

/* also, return them to caller. */
*wp=w; *hp=h;
}


/* this works out how wide each part of the tree will have to be
 * to fit all the text for each of the names, etc. (Obviously the
 * height of the tree only depends on the maximum node depth.)
 */
void treegrok()
{
int w,h;

treegrok_main(&root,&w,&h);

/* add border */
w+=2*BORDER_SIZE;
h+=2*BORDER_SIZE;

pic_width=w;
pic_height=h;
}


#ifdef DEBUG
void treeprint(struct node_tag *nptr,int indent)
{
int f;

for(f=0;f<indent;f++) putchar('-');
printf("%s (%dx%d)\n",nptr->t,nptr->sw,nptr->sh);

for(f=0;f<nptr->knum;f++)
  treeprint(*(nptr->kids+f),indent+2);
}
#endif	/* DEBUG */


/* drawing ops.
 * perhaps these belong in another file. XXX?
 */

void drawrect(int x1,int y1,int x2,int y2,int col)
{
int y;

memset(pic+y1*pic_width+x1,col,x2-x1+1);	/* top line */

for(y=y1+1;y<=y2-1;y++)
  {
  pic[y*pic_width+x1]=col;
  pic[y*pic_width+x2]=col;
  }

memset(pic+y2*pic_width+x1,col,x2-x1+1);	/* bottom line */
}


void drawrectfill(int x1,int y1,int x2,int y2,int col)
{
int y;

for(y=y1;y<=y2;y++)
  memset(pic+y*pic_width+x1,col,x2-x1+1);
}


/* drawline is a slightly hacked version of:
 *
 * digline: draw digital line from (x1,y1) to (x2,y2),
 * calling a user-supplied procedure at each pixel.
 * Does no clipping.  Uses Bresenham's algorithm.
 *
 * Paul Heckbert	3 Sep 85
 *
 * ...which is from "Graphics Gems", Academic Press, 1990.
 */

/* absolute value of a */
#define ABS(a)		(((a)<0) ? -(a) : (a))

/* take binary sign of a, either -1, or 1 if >= 0 */
#define SGN(a)		(((a)<0) ? -1 : 1)

/* I've left the indentation as it was, mainly 'cos I can't be bothered
 * to fix it :-)
 */
void drawline(int x1,int y1,int x2,int y2,int col)
{
    int d, x, y, ax, ay, sx, sy, dx, dy;

    dx = x2-x1;  ax = ABS(dx)<<1;  sx = SGN(dx);
    dy = y2-y1;  ay = ABS(dy)<<1;  sy = SGN(dy);

    x = x1;
    y = y1;
    if (ax>ay) {		/* x dominant */
	d = ay-(ax>>1);
	for (;;) {
	    pic[y*pic_width+x]=col;
	    if (x==x2) return;
	    if (d>=0) {
		y += sy;
		d -= ax;
	    }
	    x += sx;
	    d += ay;
	}
    }
    else {			/* y dominant */
	d = ax-(ay>>1);
	for (;;) {
	    pic[y*pic_width+x]=col;
	    if (y==y2) return;
	    if (d>=0) {
		x += sx;
		d -= ay;
	    }
	    y += sy;
	    d += ax;
	}
    }
}


void drawtext(int ox,int oy,char *text,int col)
{
int len=strlen(text);
int f,x,y,c;
char *ptr;	/* pointer to font data for given char */

for(f=0;f<len;f++)
  {
  /* map non-printable and non-ascii to underscore */
  c=text[f];
  if(c<32 || c>126) c='_';
  
  for(y=0;y<FONT_YDIM;y++)
    {
    ptr=font[(c-32)*FONT_YDIM+y];
    for(x=0;x<FONT_XDIM;x++)
      if(ptr[x]!=' ')
        pic[(oy+y)*pic_width+ox+FONT_XDIM*f+x]=col;
    }
  }
}



/* this routine is called from writepgm if we really want to output
 * an RLE targa file instead of pgm. Being compressed, even with crappy
 * old RLE, it'll be much smaller.
 */
void writetga(unsigned char *bytemap,int width,int height)
{
int f,x,y,c;
int pal[8];
int rle_count,old;

/* first, output the header */
putchar(0);			/* 0-len id field */
putchar(1);			/* do have palette */
putchar(9);			/* type 9 - 8-bit, RLE compressed */
putchar(0); putchar(0);		/* palette starts at entry 0 */
putchar(8); putchar(0);		/* palette is 8 entries long */
				/* (only 5 used right now, though) */
putchar(24);			/* palette entry depth (3*8-bit) */
putchar(0); putchar(0); putchar(0); putchar(0);		/* origin (0,0) */
putchar(width&255);  putchar(width/256);	/* width */
putchar(height&255); putchar(height/256);	/* height */
putchar(8);			/* bpp */
putchar(0);			/* 'desc' */

/* now the palette */
for(f=0;f<3;f++) putchar(pal[0]=grey_bg);
for(f=0;f<3;f++) putchar(pal[1]=grey_textfg);
for(f=0;f<3;f++) putchar(pal[2]=grey_textbg);
for(f=0;f<3;f++) putchar(pal[3]=grey_boxline);
for(f=0;f<3;f++) putchar(pal[4]=grey_line);
/* remaining ones unused for now */
for(f=0;f<3;f++) putchar(pal[5]=0);
for(f=0;f<3;f++) putchar(pal[6]=0);
for(f=0;f<3;f++) putchar(pal[7]=0);

/* finally, output the image.
 *
 * if after reading this code you can't believe tga's RLE works this way,
 * you have my sympathies. It's pretty pessimal even for RLE. :-( But, it's
 * more widely supported than the slightly bizarre PCX, so I'm sticking with
 * it, bogus or not. Anyway, even this rather suspect RLE (two bytes used
 * whether an RLE run or not!) ends up much smaller than pgm.
 */

for(y=height-1;y>=0;y--)
  {
  rle_count=0; old=-1;
  for(x=0;x<width;x++)
    {
    /* match pixel to palette */
    c=bytemap[y*width+x];
    for(f=0;f<8;f++)
      if(c==pal[f])
        {
        c=f;
        break;
        }
    
    /* mapped colour is in c, then */
    if(c==old && rle_count<128)
      rle_count++;
    else
      {
      if(old>=0)
        {
        if(rle_count>1)
          putchar(127+rle_count);
        else
          putchar(0);
        putchar(old);
        }
      rle_count=1;
      old=c;
      }
    }
  
  /* deal with remaining output */
  if(rle_count>1)
    putchar(127+rle_count);
  else
    putchar(0);
  putchar(old);
  }
}



/* (x,y) should point to top-centre of box.
 * this routine is fairly icky.
 */
void writepgm_main(struct node_tag *nptr,int x,int y)
{
int f,cx,cy,xtmp;

/* draw our own box */
xtmp=x-nptr->w/2;
/* outline */
drawrect(xtmp,y,xtmp+nptr->w-1,y+nptr->h-1,grey_boxline);
/* interior */
drawrectfill(xtmp+1,y+1,xtmp+nptr->w-2,y+nptr->h-2,grey_textbg);

/* draw the text */
drawtext(xtmp+1+BOX_MARGIN,y+1+BOX_MARGIN,nptr->t,grey_textfg);

/* draw lines to children, and draw their boxes */
cx=x-nptr->cw/2; cy=y+nptr->h+CHILD_VERT_GAP;
for(f=0;f<nptr->knum;f++)
  {
  /* make xtmp x pos of middle of child box */
  xtmp=cx+(*(nptr->kids+f))->sw/2;		/* yuck :-( */
  drawline(x,y+nptr->h,xtmp,cy,grey_line);	/* draw connecting line */
  writepgm_main(*(nptr->kids+f),xtmp,cy);
  cx+=(*(nptr->kids+f))->sw+CHILD_HORIZ_GAP;
  }
}


/* draw the tree in a bytemap and write it to a pgm.
 * the structure is, obviously, similar to treegrok; this is just more
 * fully fleshed-out, so to speak.
 */
void writepgm()
{
/* this may take a while (and it may be a HUGE file) so warn them so
 * they can kill us off if it sounds too much.
 */
fprintf(stderr,"treetopgm: generating %dx%d image (%dk)\n",
	pic_width,pic_height,(pic_width*pic_height+1023)/1024);

pic=xmalloc(pic_width*pic_height);

/* blank out image */
memset(pic,grey_bg,pic_width*pic_height);

writepgm_main(&root,pic_width/2,BORDER_SIZE);

/* finally! write the pgm and free the pic */
if(output_tga && pic_width>65535)
  {
  fprintf(stderr,"treetopgm: too wide for TGA, using PGM!\n");
  output_tga=0;
  }

if(output_tga)
  writetga(pic,pic_width,pic_height);
else
  if(output_mrf)
    writemrf(pic,pic_width,pic_height);
  else
    {
    printf("P5\n%d %d\n255\n",pic_width,pic_height);
    fwrite(pic,pic_width,pic_height,stdout);
    }

free(pic);
}



int main(int argc,char *argv[])
{
/* this is a little crude, but using getopt seems like overkill somehow */
if(argc==2 && argv[1][0]=='-' && strlen(argv[1])==2)
  switch(argv[1][1])
    {
    case 't': output_tga=1; break;
    case 'm': output_mrf=1; break;
    }

parsetree();	/* parse 'find' list into tree structure */
treegrok();	/* work out sizes */
writepgm();	/* construct the pgm and write it */

/* should destroy tree here, really, but... why bother? :-) */

exit(0);
}
